/* 

Pipe Mixer for ESS ISA cards by Lesha Bogdanow
Copyright (C) 1999-2000  Lesha Bogdanow

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#define USE_OS2_TOOLKIT_HEADERS
#define INCL_DOSPROCESS
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <os2.h>
#include <sys\hw.h>

#include "ioctl90.h"
#include "PipeMixPvt.h"

const char szLogo[]="PipeMixer demon for ESS ISA cards. v.0.05";

int	BasePort=0;
int	LockRate=0;
int	AccessedPort=0;

MIXSTRUCT DAC={2,0,0};		// Released initially
MIXSTRUCT Master={0,0,0};
MIXSTRUCT MIDI={0,0,0};		// IOCTL90: Aux
MIXSTRUCT CD={0,0,0};		// IOCTL90: CD
MIXSTRUCT Line={0,0,0};
MIXSTRUCT Mic={0,100,100};	// Dummy
MIXSTRUCT ADC={2,0,0};		// Released initially
MIXSTRUCT RecMux={2,0,0};	// Released initially

int	LockPriority=PRTYC_TIMECRITICAL;
TID	LockTID=0;

static void SetRegister(unsigned char idx,unsigned char data) {
   if (BasePort) {
      if (BasePort!=AccessedPort) {
         _portaccess(BasePort,BasePort+1);
         AccessedPort=BasePort;
         }
      _outp8(BasePort+4,idx);
      _outp8(BasePort+5,data);
      }
   }

static unsigned char GetRegister(unsigned char idx) {
   if (BasePort) {
      if (BasePort!=AccessedPort) {
         _portaccess(BasePort,BasePort+1);
         AccessedPort=BasePort;
         }
      _outp8(BasePort+4,idx);
      return _inp8(BasePort+5);
      }
   return 0;
   }

// Reset mixer
static void MixerReset() {
   SetRegister(0,0);
   }

static void Limit100(MIXSTRUCT *ms) {
   if (ms->VolumeL>100) ms->VolumeL=100;
   if (ms->VolumeR>100) ms->VolumeR=100;
   }
static inline int Scale(v,lim) {
   return (v*100)/(lim-1);
   }
static void _GetInp(MIXSTRUCT *ms, int r) {
   unsigned char t;

   t=GetRegister(r);
   ms->VolumeL=Scale(t>>4,16);
   ms->VolumeR=Scale(t&0xF,16);
   Limit100(ms);
   }

// Get hardware state at startup
static void GetState() {
   unsigned char t;

   t=GetRegister(0x1C);
   if (!(t&2)) RecMux.VolumeL=I90SRC_MIC;
   else if (!(t&4)) RecMux.VolumeL=I90SRC_CD;
   else if (!(t&1)) RecMux.VolumeL=I90SRC_LINE;
   else RecMux.VolumeL=I90SRC_AUX;

   _GetInp(&Master,0x32);
   _GetInp(&DAC,0x14);
   if (!(DAC.VolumeL+DAC.VolumeR)) DAC.Mute|=1;
   _GetInp(&MIDI,0x36);
   if (!(MIDI.VolumeL+MIDI.VolumeR)) MIDI.Mute|=1;
   _GetInp(&CD,0x38);
   if (!(CD.VolumeL+CD.VolumeR)) CD.Mute|=1;
   _GetInp(&Line,0x3E);
   if (!(Line.VolumeL+Line.VolumeR)) Line.Mute|=1;
   _GetInp(&ADC,0x1A);
   if (ADC.VolumeR>ADC.VolumeL) ADC.VolumeL=ADC.VolumeR;
   ADC.VolumeR=0;
   if (!ADC.VolumeL) ADC.Mute|=1;
   return;
   }

static inline int RScale(v,lim) {
   int t;
   t=(v*(lim-1))/100;
   if (t>(lim-1)) t=lim-1;
   return t;
   }

static void _SetInp(MIXSTRUCT *ms, int r) {
   int lv,rv;

   lv=ms->VolumeL;
   rv=ms->VolumeR;
   if (ms->Mute&1) {
      lv=0;
      rv=0;
      }
   SetRegister(r,(RScale(lv,16)<<4)|RScale(rv,16));
   return;   
   }

static void SetMaster() {
   _SetInp(&Master,0x32);
   return;
   }
static void SetDAC() {
   _SetInp(&DAC,0x14);
   }
static void SetMIDI() {
   _SetInp(&MIDI,0x36);
   }
static void SetCD() {
   _SetInp(&CD,0x38);
   }
static void SetLine() {
   _SetInp(&Line,0x3E);
   }

static void SetADC() {
   MIXSTRUCT ms;

   ms=ADC;
   ms.VolumeR=ms.VolumeL;
   _SetInp(&ms,0x1A);
   }

/* TODO: Input sel, tone */
static void SetRecMux() {
   unsigned char t;

   if (RecMux.VolumeL&I90SRC_MIC) t=0;
   else if (RecMux.VolumeL&I90SRC_CD) t=2;
   else if (RecMux.VolumeL&I90SRC_LINE) t=6;
   else t=7;
   SetRegister(0x1C,t);
   }

VOID APIENTRY LockThread(ULONG Dummy) {
   DosSetPriority(PRTYS_THREAD, LockPriority, 0, 0);

   while(1) {
      DosSleep(LockRate);
      SetMaster();
      SetDAC();
      SetADC();
      SetRecMux();
      };
   }

// Initialize mixer. Return TRUE if Ok
int MixerInit(int argc, char *argv[]) {
   int i,Reset;

   Reset=FALSE;
   for (i=1;i<argc;i++) {
     if (!stricmp(argv[i],"-b"))  {
        i++;
        if (i==argc) return FALSE;		// Syntax error
        BasePort=strtoul(argv[i],NULL,16)&0xFFFF;
        }
     else if (!stricmp(argv[i],"-l"))  {
        i++;
        if (i==argc) return FALSE;		// Syntax error
        LockRate=atoi(argv[i]);
        }
     else if (!stricmp(argv[i],"-r")) Reset=TRUE;
     else return FALSE;				// Syntax error
     }
   if (!BasePort) return FALSE;
   GetState();
   if (LockRate) DosCreateThread(&LockTID,LockThread,0,0,8192);
   ApiLevel=2;
   if (Debug) printf("API Level: %d, Base port: %x, Lock rate: %d\n",ApiLevel,
                      BasePort,LockRate);
   if (Reset) {
      if (Debug) printf("Resetting mixer.\n");
      MixerReset();
      }
   return TRUE;
   }

// Deinitialize mixer
void MixerClose() {
   if (LockTID) DosKillThread(LockTID);
   }

/* Process a command. Buff is an output buffer, the function number and
   a trailing space are already there. tokens - number of command tokens
   cmd - command, p1, p2, p3 - parameters. If some of them are not specified
   then default values (p1=100, p2=p1, p3=0) are supplied
*/
int ProcessCmd(char *Buff, int tokens, int cmd, int p1, int p2, int p3) {
   MIXSTRUCT MixStruct;

   if (((cmd&0xF0)==0x40)||(cmd==0x01)) {	// Set command
      MixStruct.VolumeL=p1;
      MixStruct.VolumeR=p2;
      MixStruct.Mute=p3;

      switch(cmd) {
         case 0x01:
            Master=MixStruct;
            SetMaster();
            break;
         case 0x42:		// Mic, dummy
            Mic=MixStruct;
            break;
         case 0x43:		// Line
            Line=MixStruct;
            SetLine();
            break;
         case 0x44:		// CD
            CD=MixStruct;
            SetCD();
            break;
         case 0x46:		// Aux
            MIDI=MixStruct;
            SetMIDI();
            break;
         case 0x4D:		// DAC
            DAC=MixStruct;
            SetDAC();
            break;
         case 0x4E:		// RecMux
            RecMux=MixStruct;
            SetRecMux();
            break;
         case 0x4F:		// ADC
            ADC=MixStruct;
            SetADC();
            break;
         default:
            strcat(Buff,REP_UNSUPPORTED);
            return TRUE;
         }
      strcat(Buff,REP_OK);
      }
   else if (((cmd&0xF0)==0x60)||(cmd==0x11)) {	// Query command
      switch(cmd) {
         case 0x11:
            MixStruct=Master;
            break;
         case 0x62:		// Mic
            MixStruct=Mic;
            break;
         case 0x63:		// Line
            MixStruct=Line;
            break;
         case 0x64:		// CD
            MixStruct=CD;
            break;
         case 0x66:		// Aux
            MixStruct=MIDI;
            break;
         case 0x6D:		// DAC
            MixStruct=DAC;
            break;
         case 0x6E:		// RecMux
            MixStruct=RecMux;
            break;
         case 0x6F:		// ADC
            MixStruct=ADC;
            break;
         default:
            strcat(Buff,REP_UNSUPPORTED);
            return TRUE;
         }
      sprintf(Buff+strlen(Buff),"%d %d %x",MixStruct.VolumeL,MixStruct.VolumeR,MixStruct.Mute);
      }
   else strcat(Buff,REP_UNSUPPORTED);
   return TRUE;
   }

