/* 

Pipe Mixer for CT1745 mixer by Lesha Bogdanow
Copyright (C) 1999-2000  Lesha Bogdanow

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#define USE_OS2_TOOLKIT_HEADERS
#define INCL_DOSPROCESS
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <os2.h>
#include <sys\hw.h>

#include "ioctl90.h"
#include "PipeMixPvt.h"

const char szLogo[]="PipeMixer demon for SB16/SB32/AWE32/AWE64 cards. v.0.05";

int	BasePort=0;
int	LockRate=0;
int	AccessedPort=0;

MIXSTRUCT DAC={2,0,0};		// Released initially
MIXSTRUCT Master={0,0,0};
MIXSTRUCT MIDI={0,0,0};		// IOCTL90: Aux
MIXSTRUCT CD={0,0,0};		// IOCTL90: CD
MIXSTRUCT Line={0,0,0};
MIXSTRUCT Mic={0,0,0};		
MIXSTRUCT Mono={0,0,0};		
MIXSTRUCT Tone={0,0,0};
MIXSTRUCT ADC={2,0,0};		// Released initially
MIXSTRUCT RecMix={2,0,0};	// Released initially

int	LockPriority=PRTYC_TIMECRITICAL;
TID	LockTID=0;

static void SetRegister(unsigned char idx,unsigned char data) {
   if (BasePort) {
      if (BasePort!=AccessedPort) {
         _portaccess(BasePort,BasePort+1);
         AccessedPort=BasePort;
         }
      _outp8(BasePort+4,idx);
      _outp8(BasePort+5,data);
      }
   }

static unsigned char GetRegister(unsigned char idx) {
   if (BasePort) {
      if (BasePort!=AccessedPort) {
         _portaccess(BasePort,BasePort+1);
         AccessedPort=BasePort;
         }
      _outp8(BasePort+4,idx);
      return _inp8(BasePort+5);
      }
   return 0;
   }


static void Limit100(MIXSTRUCT *ms) {
   if (ms->VolumeL>100) ms->VolumeL=100;
   if (ms->VolumeR>100) ms->VolumeR=100;
   }

static inline int Scale(v,lim) {
   return (v*100)/(lim-1);
   }
static void _GetInp(MIXSTRUCT *ms, int r) {
   unsigned char t;

   t=GetRegister(r)>>3;
   ms->VolumeL=Scale(t,32);
   t=GetRegister(r+1)>>3;
   ms->VolumeR=Scale(t,32);
   Limit100(ms);
   }
// Reset mixer
static void MixerReset() {
   SetRegister(0,0);
   }

// Get hardware state at startup
static void GetState() {
   unsigned char t;

   t=GetRegister(0x3D)|GetRegister(0x3E);
   RecMix.VolumeL=0;
   if (t&1) RecMix.VolumeL|=I90SRC_MIC;
   if (t&6) RecMix.VolumeL|=I90SRC_CD;
   if (t&0x18) RecMix.VolumeL|=I90SRC_LINE;
   if (t&0x60) RecMix.VolumeL|=I90SRC_AUX;

   t=(GetRegister(0x30)>>3)|((GetRegister(0x41)&0xC0)>>1);
   Master.VolumeL=Scale(t,128);
   t=(GetRegister(0x31)>>3)|((GetRegister(0x42)&0xC0)>>1);
   Master.VolumeR=Scale(t,128);
   Limit100(&Master);
   if (!(Master.VolumeL+Master.VolumeR)) Master.Mute|=1;

   t=GetRegister(0x3C);
   _GetInp(&DAC,0x32);
   if (!(DAC.VolumeL+DAC.VolumeR)) DAC.Mute|=1;
   _GetInp(&MIDI,0x34);
   if (!(MIDI.VolumeL+MIDI.VolumeR)) MIDI.Mute|=1;
   _GetInp(&CD,0x36);
   if (!(t&6)) CD.Mute|=1;
   _GetInp(&Line,0x38);
   if (!(t&0x18)) Line.Mute|=1;
   if (!(t&1)) Mic.Mute|=1;
   t=GetRegister(0x3A)>>3;
   Mic.VolumeL=Scale(t,32);
   Mic.VolumeR=Mic.VolumeL;
   Limit100(&Mic);
   t=GetRegister(0x3B)>>6;
   Mono.VolumeL=Scale(t,4);
   Limit100(&Mono);
   if (!Mono.VolumeL) Mono.Mute|=1;
   t=GetRegister(0x3F)>>6;
   ADC.VolumeL=Scale(t,4);
   Limit100(&ADC);
   if (!ADC.VolumeL) ADC.Mute|=1;
   t=GetRegister(0x44)>>4;
   Tone.VolumeR=Scale(t,16);
   t=GetRegister(0x46)>>4;
   Tone.VolumeL=Scale(t,16);
   Limit100(&Tone);
   return;
   }

/* NO SET FUNCTION IS FINISHED */
static inline int RScale(v,lim) {
   int t;
   t=(v*(lim-1))/100;
   if (t>(lim-1)) t=lim-1;
   return t;
   }

static void SetMaster() {
   unsigned char t;
   
   t=RScale(Master.VolumeL,128);
   SetRegister(0x30,(t&0x1F)<<3);
   SetRegister(0x41,(t&0x60)<<1);
   t=RScale(Master.VolumeR,128);
   SetRegister(0x31,(t&0x1F)<<3);
   SetRegister(0x42,(t&0x60)<<1);
   return;
   }

// me = TRUE - emulate mute.
static void _SetInp(MIXSTRUCT *ms, int r,int me) {
   int lv,rv;

   lv=ms->VolumeL;
   rv=ms->VolumeR;
   if ((me)&&(ms->Mute&1)) {
      lv=0;
      rv=0;
      }
   SetRegister(r,RScale(lv,32)<<3);
   SetRegister(r+1,RScale(rv,32)<<3);
   return;   
   }
static void SetDAC() {
   _SetInp(&DAC,0x32,TRUE);
   }
static void SetMIDI() {
   _SetInp(&MIDI,0x34,TRUE);
   }
static void _SetInpM(MIXSTRUCT *ms, int r, unsigned char mask) {
   unsigned char t;

   _SetInp(ms,r,FALSE);
   t=GetRegister(0x3C);
   if (ms->Mute&1) t&=~mask;
   else t|=mask;
   SetRegister(0x3C,t);
   return;
   }
static void SetCD() {
   _SetInpM(&CD,0x36,6);
   }
static void SetLine() {
   _SetInpM(&CD,0x38,0x18);
   }

static void SetMic() {
   unsigned char t;

   SetRegister(0x3A,RScale(Mic.VolumeL,32)<<3);
   t=GetRegister(0x3C);
   if (Mic.Mute&1) t&=0xFE;
   else t|=1;
   SetRegister(0x3C,t);
   return;
   }

static void SetMono() {
   if (Mono.Mute&1) SetRegister(0x3B,0);
   else SetRegister(0x3B,RScale(Mono.VolumeL,4)<<6);
   return;
   }

static void SetADC() {
   unsigned char t;
   if (ADC.Mute&1) t=0;
   else t=RScale(ADC.VolumeL,4)<<6;
   SetRegister(0x3F,t);
   SetRegister(0x40,t);
   }

/* TODO: Input sel, tone */
static void SetRecMix() {
   unsigned char tl,tr;

   tl=0;
   tr=0;
   if (RecMix.VolumeL&I90SRC_MIC) {
      tl|=1;
      tr|=1;
      }
   if (RecMix.VolumeL&I90SRC_CD) {
      tl|=4;
      tr|=2;
      }
   if (RecMix.VolumeL&I90SRC_LINE) {
      tl|=0x10;
      tr|=0x08;
      }
   if (RecMix.VolumeL&I90SRC_AUX) {
      tl|=0x40;
      tr|=0x20;
      }
   SetRegister(0x3D,tl);
   SetRegister(0x3E,tr);
   }

static void SetTone() {
   unsigned char t;

   t=RScale(Tone.VolumeL,16)<<4;
   SetRegister(0x46,t);
   SetRegister(0x47,t);
   t=RScale(Tone.VolumeR,16)<<4;
   SetRegister(0x44,t);
   SetRegister(0x45,t);
   }

VOID APIENTRY LockThread(ULONG Dummy) {
   DosSetPriority(PRTYS_THREAD, LockPriority, 0, 0);

   while(1) {
      DosSleep(LockRate);
      SetMaster();
      SetDAC();
      SetADC();
      SetRecMix();
      };
   }

// Initialize mixer. Return TRUE if Ok
int MixerInit(int argc, char *argv[]) {
   int i,Reset;

   Reset=FALSE;
   for (i=1;i<argc;i++) {
     if (!stricmp(argv[i],"-b"))  {
        i++;
        if (i==argc) return FALSE;		// Syntax error
        BasePort=strtoul(argv[i],NULL,16)&0xFFFF;
        }
     else if (!stricmp(argv[i],"-l"))  {
        i++;
        if (i==argc) return FALSE;		// Syntax error
        LockRate=atoi(argv[i]);
        }
     else if (!stricmp(argv[i],"-r")) Reset=TRUE;
     else return FALSE;				// Syntax error
     }
   if (!BasePort) return FALSE;
   GetState();
   if (LockRate) DosCreateThread(&LockTID,LockThread,0,0,8192);
   ApiLevel=2;
   if (Debug) printf("API Level: %d, Base port: %x, Lock rate: %d\n",ApiLevel,
                      BasePort,LockRate);
   if (Reset) {
      if (Debug) printf("Resetting mixer.\n");
      MixerReset();
      }
   return TRUE;
   }

// Deinitialize mixer
void MixerClose() {
   if (LockTID) DosKillThread(LockTID);
   }

/* Process a command. Buff is an output buffer, the function number and
   a trailing space are already there. tokens - number of command tokens
   cmd - command, p1, p2, p3 - parameters. If some of them are not specified
   then default values (p1=100, p2=p1, p3=0) are supplied
*/
int ProcessCmd(char *Buff, int tokens, int cmd, int p1, int p2, int p3) {
   MIXSTRUCT MixStruct;

   if (((cmd&0xF0)==0x40)||(cmd==0x01)) {	// Set command
      MixStruct.VolumeL=p1;
      MixStruct.VolumeR=p2;
      MixStruct.Mute=p3;

      switch(cmd) {
         case 0x01:
            Master=MixStruct;
            SetMaster();
            break;
         case 0x40:
            Mono=MixStruct;
            SetMono();
            break;
         case 0x42:		// Mic
            Mic=MixStruct;
            SetMic();
            break;
         case 0x43:		// Line
            Line=MixStruct;
            SetLine();
            break;
         case 0x44:		// CD
            CD=MixStruct;
            SetCD();
            break;
         case 0x46:		// Aux
            MIDI=MixStruct;
            SetMIDI();
            break;
         case 0x4B:		// Tone
            Tone=MixStruct;
            SetTone();
            break;
         case 0x4D:		// DAC
            DAC=MixStruct;
            SetDAC();
            break;
         case 0x4E:		// RecMix
            RecMix=MixStruct;
            SetRecMix();
            break;
         case 0x4F:		// ADC
            ADC=MixStruct;
            SetADC();
            break;
         default:
            strcat(Buff,REP_UNSUPPORTED);
            return TRUE;
         }
      strcat(Buff,REP_OK);
      }
   else if (((cmd&0xF0)==0x60)||(cmd==0x11)) {	// Query command
      switch(cmd) {
         case 0x11:
            MixStruct=Master;
            break;
         case 0x60:
            MixStruct=Mono;
            break;
         case 0x62:		// Mic
            MixStruct=Mic;
            break;
         case 0x63:		// Line
            MixStruct=Line;
            break;
         case 0x64:		// CD
            MixStruct=CD;
            break;
         case 0x66:		// Aux
            MixStruct=MIDI;
            break;
         case 0x6B:		// Tone
            MixStruct=Tone;
            MixStruct.Mute=3;
            break;
         case 0x6D:		// DAC
            MixStruct=DAC;
            break;
         case 0x6E:		// RecMix
            MixStruct=RecMix;
            MixStruct.Mute|=4;
            break;
         case 0x6F:		// ADC
            MixStruct=ADC;
            break;
         default:
            strcat(Buff,REP_UNSUPPORTED);
            return TRUE;
         }
      sprintf(Buff+strlen(Buff),"%d %d %x",MixStruct.VolumeL,MixStruct.VolumeR,MixStruct.Mute);
      }
   else strcat(Buff,REP_UNSUPPORTED);
   return TRUE;
   }

