/* 

LBMix - A mixer for IOCTL90 API by Lesha Bogdanow
Copyright (C) 1999-2000  Lesha Bogdanow

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/


// This file contain very generic routines related to PM and GPI


#define USE_OS2_TOOLKIT_HEADERS
#define INCL_WIN
#define INCL_HELP
#define INCL_WINHEAP
#define INCL_WINMENU		   /* For menu control function */
#define INCL_WINDIALOGS
#define INCL_WINMESSAGEMGR
#define INCL_DOSPROCESS			   /* For init function */
#define INCL_GPI
#define INCL_BASE
#define INCL_ERRORS

#include <stddef.h>
#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include <os2.h>
#include "GUITools.h"


static int clrmix(int c1, int i1, int c2, int i2) {
   int rc;

   rc=(i1*(c1&0xFF)+i2*(c2&0xFF))/(i1+i2);
   if (rc>0xFF) rc=0xFF;
   return rc;
   }

// Mix RGB colors c1 and c2 in i1 to i2 proportion
LONG RGBMix(LONG c1, int i1, LONG c2, int i2) {
   LONG rc;

   rc=clrmix(c1,i1,c2,i2);
   rc|=clrmix(c1>>8,i1,c2>>8,i2)<<8;
   rc|=clrmix(c1>>16,i1,c2>>16,i2)<<16;
   return rc;
   }

static int PPColors1[]={PP_FOREGROUNDCOLOR,
                        PP_BACKGROUNDCOLOR,
                        PP_HILITEFOREGROUNDCOLOR,
                        PP_HILITEBACKGROUNDCOLOR,
                        PP_DISABLEDFOREGROUNDCOLOR,
                        PP_DISABLEDBACKGROUNDCOLOR,
                        PP_BORDERCOLOR,
                        PP_ACTIVECOLOR,
                        PP_INACTIVECOLOR,
                        PP_ACTIVETEXTFGNDCOLOR,
                        PP_ACTIVETEXTBGNDCOLOR,
                        PP_INACTIVETEXTFGNDCOLOR,
                        PP_INACTIVETEXTBGNDCOLOR,
                        PP_SHADOW,
                        PP_MENUFOREGROUNDCOLOR,
                        PP_MENUBACKGROUNDCOLOR,
                        PP_MENUHILITEFGNDCOLOR,
                        PP_MENUHILITEBGNDCOLOR,
                        PP_MENUDISABLEDFGNDCOLOR,
                        PP_MENUDISABLEDBGNDCOLOR,
                        PP_SHADOWTEXTCOLOR,
                        PP_SHADOWHILITEFGNDCOLOR,
                        PP_SHADOWHILITEBGNDCOLOR,
                        PP_ICONTEXTBACKGROUNDCOLOR,
                        PP_BORDERLIGHTCOLOR,
                        PP_BORDERDARKCOLOR,
                        PP_BORDER2COLOR,
                        PP_BORDER2LIGHTCOLOR,
                        PP_BORDER2DARKCOLOR,
                        PP_BORDERDEFAULTCOLOR,
                        PP_FIELDBACKGROUNDCOLOR,
                        PP_BUTTONBACKGROUNDCOLOR,
                        PP_BUTTONBORDERLIGHTCOLOR,
                        PP_BUTTONBORDERDARKCOLOR,
                        PP_ARROWCOLOR,
                        PP_ARROWBORDERLIGHTCOLOR,
                        PP_ARROWBORDERDARKCOLOR,
                        PP_ARROWDISABLEDCOLOR,
                        PP_CHECKLIGHTCOLOR,
                        PP_CHECKMIDDLECOLOR,
                        PP_CHECKDARKCOLOR,
                        PP_PAGEFOREGROUNDCOLOR,
                        PP_PAGEBACKGROUNDCOLOR,
                        PP_MAJORTABFOREGROUNDCOLOR,
                        PP_MAJORTABBACKGROUNDCOLOR,
                        PP_MINORTABFOREGROUNDCOLOR,
                        PP_MINORTABBACKGROUNDCOLOR};
static int PPColors2[]={PP_FOREGROUNDCOLORINDEX,
                        PP_BACKGROUNDCOLORINDEX,
                        PP_HILITEFOREGROUNDCOLORINDEX,
                        PP_HILITEBACKGROUNDCOLORINDEX,
                        PP_DISABLEDFOREGROUNDCOLORINDEX,
                        PP_DISABLEDBACKGROUNDCOLORINDEX,
                        PP_BORDERCOLORINDEX,
                        PP_ACTIVECOLORINDEX,
                        PP_INACTIVECOLORINDEX,
                        PP_ACTIVETEXTFGNDCOLORINDEX,
                        PP_ACTIVETEXTBGNDCOLORINDEX,
                        PP_INACTIVETEXTFGNDCOLORINDEX,
                        PP_INACTIVETEXTBGNDCOLORINDEX,
                        0,
                        PP_MENUFOREGROUNDCOLORINDEX,
                        PP_MENUBACKGROUNDCOLORINDEX,
                        PP_MENUHILITEFGNDCOLORINDEX,
                        PP_MENUHILITEBGNDCOLORINDEX,
                        PP_MENUDISABLEDFGNDCOLORINDEX,
                        PP_MENUDISABLEDBGNDCOLORINDEX,
                        PP_SHADOWTEXTCOLORINDEX,
                        PP_SHADOWHILITEFGNDCOLORINDEX,
                        PP_SHADOWHILITEBGNDCOLORINDEX,
                        PP_ICONTEXTBACKGROUNDCOLORINDEX,
			0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

static PPRESPARAMS AddPP(PPRESPARAMS pp, int id, int cb, void *data) {
   PARAM *p;
   int idx;
   
   idx=pp->cb;
   pp->cb+=2*sizeof(ULONG)+cb;
   pp=realloc(pp,pp->cb+sizeof(ULONG));
   p=(PARAM *)((void*)(&(pp->aparam[0]))+idx);
   p->id=id;
   p->cb=cb;
   memcpy(p->ab,data,cb);
   return pp;
   }

// Query presentation parameters of a window. Return NULL if no specific
// presentation parameters are present
// returns NULL if hwnd==NULLHANDLE
PPRESPARAMS QueryWinPP(HWND hwnd) {
   PPRESPARAMS pp;
   LONG lColor;
   char FName[256];
   int i;

   if (!hwnd) return NULL;
   pp=malloc(sizeof(ULONG));
   pp->cb=0;
   for (i=0;i<(sizeof(PPColors1)/sizeof(int));i++) {
      if (WinQueryPresParam(hwnd,PPColors1[i],PPColors2[i],NULL,sizeof(lColor),
                              &lColor,QPF_ID2COLORINDEX|QPF_NOINHERIT)) {
         pp=AddPP(pp,PPColors1[i],sizeof(ULONG),&lColor);
         }
      }
   i=WinQueryPresParam(hwnd,PP_FONTNAMESIZE,0,NULL,256,FName,QPF_NOINHERIT);
   if (i) pp=AddPP(pp,PP_FONTNAMESIZE,i,FName);
   if (!pp->cb) {
      free(pp);
      pp=NULL;
      }
   return pp;
   }

/* Strip specified parameter from the PRESPARAMS structure. Returns NULL
   if pp is NULL or the parameter removed was the only one */
PPRESPARAMS StripPP(PPRESPARAMS pp, ULONG id) {
   int left,sz,icnt;
   PARAM *p;

   if (!pp) return NULL;
   icnt=pp->cb;
   left=icnt;
   p=&pp->aparam[0];
   while (left>(sizeof(ULONG)*2)) {
      if (left<(sizeof(ULONG)*2+p->cb)) break;
      sz=p->cb+sizeof(ULONG)*2;
      if (p->id==id) {
         if (left>sz) memmove(p,(void*)p+sz,left-sz);
         left-=sz;
         pp->cb-=sz;
         }
      else {
         left-=sz;
         p=(PARAM*)((void *)p+sz);
         }
      }
   if (icnt==pp->cb) return pp;
   if (!pp->cb) {
      free(pp);
      return NULL;
      }
   return realloc(pp,pp->cb+sizeof(ULONG));
   }

// Set presentation parameters for a window.
// NOP if either hwnd or pp is NULL
void SetWinPP(HWND hwnd, PPRESPARAMS pp) {
   int left,sz;
   PARAM *p;

   if ((!pp)||(!hwnd)) return;
   left=pp->cb;
   p=&pp->aparam[0];
   while (left>(sizeof(ULONG)*2)) {
      if (left<(sizeof(ULONG)*2+p->cb)) break;
      WinSetPresParam(hwnd,p->id,p->cb,p->ab);
      sz=p->cb+sizeof(ULONG)*2;
      left-=sz;
      p=(PARAM*)((void *)p+sz);
      }
   }

// Load binary data from profile. Return 1 if Ok, 0 if failed
int  LoadPrfBin(HINI hini,const char *app,const char *key, void *data, ULONG l) {
   int rc;
   ULONG len;

   rc=PrfQueryProfileSize(hini,(char *)app,(char *)key,&len);
   if ((!rc)||(len!=l)) return 0;
   rc=PrfQueryProfileData(hini,(char *)app,(char *)key,data,&len);
   if ((!rc)||(len!=l))  return 0;
   return 1;
   }


// Allocate memory and load data from profile. Return NULL if
// no data loaded
void *LoadPrfData(HINI hini,const char *app,const char *key, ULONG *len) {
   char *data;
   int rc;

   rc=PrfQueryProfileSize(hini,(char *)app,(char *)key,len);
   if ((!rc)||(!*len)) {
      *len=0;
      return NULL;
      }
   data=malloc(*len);
   rc=PrfQueryProfileData(hini,(char *)app,(char *)key,data,len);
   if ((!rc)||(!*len)) {
      free(data);
      *len=0;
      return NULL;
      }
   return data;
   }

// Save presentation parameters to profile
void SavePP(HINI hini,const char *app,const char *key, PPRESPARAMS pp) {
   if (pp) PrfWriteProfileData(hini,(char *)app,(char *)key,pp,pp->cb+sizeof(ULONG));
   else PrfWriteProfileData(hini,(char *)app,(char *)key,NULL,0);
   }

// Allocate memory and load presentation parameters from profile
PPRESPARAMS LoadPP(HINI hini,const char *app,const char *key) {
   ULONG len;
   PPRESPARAMS data;

   data=LoadPrfData(hini,app,key,&len);
   if (!data) return NULL;
   if (len<sizeof(ULONG)) {
      free(data);
      return NULL;
      }
   if ((data->cb+sizeof(ULONG))>len)  {
      free(data);
      return NULL;
      }
   return data;
   }
